# Requirements: diffusers==0.25.1.
# Reference: https://github.com/InstantID/InstantID/blob/main/gradio_demo/requirements.txt
REPO = None
FACE_ANALYSIS_ROOT = None
MODEL = "wangqixun/YamerMIX_v8"
VARIANT = None
CUSTOM_PIPELINE = "pipeline_stable_diffusion_xl_instantid"
SCHEDULER = "EulerAncestralDiscreteScheduler"
LORA = None
CONTROLNET = "InstantX/InstantID"
STEPS = 30
PROMPT = "film noir style, ink sketch|vector, male man, highly detailed, sharp focus, ultra sharpness, monochrome, high contrast, dramatic shadows, 1940s style, mysterious, cinematic"
NEGATIVE_PROMPT = "ugly, deformed, noisy, blurry, low contrast, realism, photorealistic, vibrant, colorful"
SEED = None
WARMUPS = 3
BATCH = 1
HEIGHT = None
WIDTH = None
INPUT_IMAGE = "https://github.com/InstantID/InstantID/blob/main/examples/musk_resize.jpeg?raw=true"
OUTPUT_IMAGE = None
EXTRA_CALL_KWARGS = """{
    "controlnet_conditioning_scale": 0.8,
    "ip_adapter_scale": 0.8
}"""
CACHE_INTERVAL = 3
CACHE_LAYER_ID = 0
CACHE_BLOCK_ID = 0

import argparse
import importlib
import inspect
import json
import os
import sys
import time

import cv2
import numpy as np
import torch
from diffusers.utils import load_image
from huggingface_hub import snapshot_download
from insightface.app import FaceAnalysis
from PIL import Image, ImageDraw

import oneflow as flow  # usort: skip
from onediffx import compile_pipe


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--repo", type=str, default=REPO)
    parser.add_argument("--face-analysis-root", type=str, default=FACE_ANALYSIS_ROOT)
    parser.add_argument("--model", type=str, default=MODEL)
    parser.add_argument("--variant", type=str, default=VARIANT)
    parser.add_argument("--custom-pipeline", type=str, default=CUSTOM_PIPELINE)
    parser.add_argument("--scheduler", type=str, default=SCHEDULER)
    parser.add_argument("--lora", type=str, default=LORA)
    parser.add_argument("--controlnet", type=str, default=CONTROLNET)
    parser.add_argument("--steps", type=int, default=STEPS)
    parser.add_argument("--prompt", type=str, default=PROMPT)
    parser.add_argument("--negative-prompt", type=str, default=NEGATIVE_PROMPT)
    parser.add_argument("--seed", type=int, default=SEED)
    parser.add_argument("--warmups", type=int, default=WARMUPS)
    parser.add_argument("--batch", type=int, default=BATCH)
    parser.add_argument("--height", type=int, default=HEIGHT)
    parser.add_argument("--width", type=int, default=WIDTH)
    parser.add_argument("--cache_interval", type=int, default=CACHE_INTERVAL)
    parser.add_argument("--cache_layer_id", type=int, default=CACHE_LAYER_ID)
    parser.add_argument("--cache_block_id", type=int, default=CACHE_BLOCK_ID)
    parser.add_argument("--extra-call-kwargs", type=str, default=EXTRA_CALL_KWARGS)
    parser.add_argument("--input-image", type=str, default=INPUT_IMAGE)
    parser.add_argument("--output-image", type=str, default=OUTPUT_IMAGE)
    parser.add_argument("--deepcache", action="store_true")
    parser.add_argument(
        "--compiler",
        type=str,
        default="oneflow",
        choices=["none", "oneflow", "compile", "compile-max-autotune"],
    )
    return parser.parse_args()


def load_pipe(
    pipeline_cls,
    model_name,
    variant=None,
    custom_pipeline=None,
    scheduler=None,
    lora=None,
    controlnet=None,
):
    extra_kwargs = {}
    if custom_pipeline is not None:
        extra_kwargs["custom_pipeline"] = custom_pipeline
    if variant is not None:
        extra_kwargs["variant"] = variant
    if controlnet is not None:
        from diffusers import ControlNetModel

        controlnet = ControlNetModel.from_pretrained(
            controlnet,
            torch_dtype=torch.float16,
        )
        extra_kwargs["controlnet"] = controlnet
    if os.path.exists(os.path.join(model_name, "calibrate_info.txt")):
        from onediff.quantization import QuantPipeline

        pipe = QuantPipeline.from_pretrained(
            pipeline_cls, model_name, torch_dtype=torch.float16, **extra_kwargs
        )
    else:
        pipe = pipeline_cls.from_pretrained(
            model_name, torch_dtype=torch.float16, **extra_kwargs
        )
    if scheduler is not None:
        scheduler_cls = getattr(importlib.import_module("diffusers"), scheduler)
        pipe.scheduler = scheduler_cls.from_config(pipe.scheduler.config)
    if lora is not None:
        pipe.load_lora_weights(lora)
        pipe.fuse_lora()
    pipe.safety_checker = None
    pipe.to(torch.device("cuda"))
    return pipe


class IterationProfiler:
    def __init__(self):
        self.begin = None
        self.end = None
        self.num_iterations = 0

    def get_iter_per_sec(self):
        if self.begin is None or self.end is None:
            return None
        self.end.synchronize()
        dur = self.begin.elapsed_time(self.end)
        return self.num_iterations / dur * 1000.0

    def callback_on_step_end(self, pipe, i, t, callback_kwargs={}):
        if self.begin is None:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.begin = event
        else:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.end = event
            self.num_iterations += 1
        return callback_kwargs


def main():
    args = parse_args()

    assert (
        args.controlnet is not None
    ), "Please set `--controlnet` to the name or path of the controlnet"
    assert (
        args.face_analysis_root is not None
    ), "Please set `--face-analysis-root` to the path of the working directory of insightface.app.FaceAnalysis"
    assert os.path.isdir(
        os.path.join(args.face_analysis_root, "models", "antelopev2")
    ), f"Please download models from https://drive.google.com/file/d/18wEUfMNohBJ4K3Ly5wpTejPfDzp-8fI8/view?usp=sharing and extract to {os.path.join(args.face_analysis_root, 'models', 'antelopev2')}"
    assert (
        args.input_image is not None
    ), "Please set `--input-image` to the path of the input image"
    assert not args.deepcache

    if args.compiler == "oneflow":
        # Patch the attention processors to make them compatible with us.
        attention_processor_path = os.path.join(
            args.repo, "ip_adapter", "attention_processor.py"
        )

        with open(attention_processor_path, "r") as f:
            content = f.read()

        if "__call__" in content:
            content = content.replace("__call__", "forward")
            with open(attention_processor_path, "w") as f:
                f.write(content)

    custom_pipeline = None
    draw_kps = None
    if args.repo is None:
        custom_pipeline = args.custom_pipeline
        from diffusers import DiffusionPipeline

        pipeline_cls = DiffusionPipeline
    else:
        sys.path.insert(0, args.repo)

        from pipeline_stable_diffusion_xl_instantid import (
            draw_kps,
            StableDiffusionXLInstantIDPipeline as pipeline_cls,
        )

    if os.path.exists(args.controlnet):
        controlnet = args.controlnet
    else:
        controlnet = snapshot_download(args.controlnet)

    app = FaceAnalysis(
        name="antelopev2",
        root=args.face_analysis_root,
        providers=["CUDAExecutionProvider", "CPUExecutionProvider"],
    )
    app.prepare(ctx_id=0, det_size=(640, 640))

    face_adapter = os.path.join(controlnet, "ip-adapter.bin")
    controlnet_path = os.path.join(controlnet, "ControlNetModel")

    pipe = load_pipe(
        pipeline_cls,
        args.model,
        variant=args.variant,
        custom_pipeline=custom_pipeline,
        scheduler=args.scheduler,
        lora=args.lora,
        controlnet=controlnet_path,
    )

    if draw_kps is None:
        draw_kps = sys.modules[pipe.__module__].draw_kps

    pipe.load_ip_adapter_instantid(face_adapter)

    height = args.height or pipe.unet.config.sample_size * pipe.vae_scale_factor
    width = args.width or pipe.unet.config.sample_size * pipe.vae_scale_factor

    if args.compiler == "none":
        pass
    elif args.compiler == "oneflow":
        pipe = compile_pipe(pipe)
    elif args.compiler in ("compile", "compile-max-autotune"):
        mode = "max-autotune" if args.compiler == "compile-max-autotune" else None
        pipe.unet = torch.compile(pipe.unet, mode=mode)
        if hasattr(pipe, "controlnet"):
            pipe.controlnet = torch.compile(pipe.controlnet, mode=mode)
        pipe.vae = torch.compile(pipe.vae, mode=mode)
    else:
        raise ValueError(f"Unknown compiler: {args.compiler}")

    input_image = load_image(args.input_image)
    input_image = input_image.resize((width, height), Image.LANCZOS)

    face_image = input_image
    face_info = app.get(cv2.cvtColor(np.array(face_image), cv2.COLOR_RGB2BGR))
    face_info = sorted(
        face_info,
        key=lambda x: (x["bbox"][2] - x["bbox"][0]) * x["bbox"][3] - x["bbox"][1],
    )[
        -1
    ]  # only use the maximum face
    face_emb = face_info["embedding"]
    face_kps = draw_kps(face_image, face_info["kps"])

    def get_kwarg_inputs():
        kwarg_inputs = dict(
            prompt=args.prompt,
            negative_prompt=args.negative_prompt,
            image_embeds=face_emb,
            image=face_kps,
            height=height,
            width=width,
            num_inference_steps=args.steps,
            num_images_per_prompt=args.batch,
            generator=None
            if args.seed is None
            else torch.Generator(device="cuda").manual_seed(args.seed),
            **(
                dict()
                if args.extra_call_kwargs is None
                else json.loads(args.extra_call_kwargs)
            ),
        )
        if args.deepcache:
            kwarg_inputs["cache_interval"] = args.cache_interval
            kwarg_inputs["cache_layer_id"] = args.cache_layer_id
            kwarg_inputs["cache_block_id"] = args.cache_block_id
        return kwarg_inputs

    # NOTE: Warm it up.
    # The initial calls will trigger compilation and might be very slow.
    # After that, it should be very fast.
    if args.warmups > 0:
        print("Begin warmup")
        for _ in range(args.warmups):
            pipe(**get_kwarg_inputs())
        print("End warmup")

    # Let"s see it!
    # Note: Progress bar might work incorrectly due to the async nature of CUDA.
    kwarg_inputs = get_kwarg_inputs()
    iter_profiler = IterationProfiler()
    if "callback_on_step_end" in inspect.signature(pipe).parameters:
        kwarg_inputs["callback_on_step_end"] = iter_profiler.callback_on_step_end
    elif "callback" in inspect.signature(pipe).parameters:
        kwarg_inputs["callback"] = iter_profiler.callback_on_step_end
    begin = time.time()
    output_images = pipe(**kwarg_inputs).images
    end = time.time()

    print("=======================================")
    print(f"Inference time: {end - begin:.3f}s")
    iter_per_sec = iter_profiler.get_iter_per_sec()
    if iter_per_sec is not None:
        print(f"Iterations per second: {iter_per_sec:.3f}")
    cuda_mem_after_used = flow._oneflow_internal.GetCUDAMemoryUsed()
    host_mem_after_used = flow._oneflow_internal.GetCPUMemoryUsed()
    print(f"CUDA Mem after: {cuda_mem_after_used / 1024:.3f}GiB")
    print(f"Host Mem after: {host_mem_after_used / 1024:.3f}GiB")
    print("=======================================")

    if args.output_image is not None:
        output_images[0].save(args.output_image)
    else:
        print("Please set `--output-image` to save the output image")


if __name__ == "__main__":
    main()



from pathlib import Path

# ComfyUI
from comfy import model_management
from folder_paths import get_input_directory

# onediff
from onediff.infer_compiler import oneflow_compile, OneflowCompileOptions
from onediff.infer_compiler.backends.oneflow.transform import torch2oflow
from onediff.optimization.quant_optimizer import quantize_model

# onediff_comfy_nodes
from .model_patcher import state_dict_hook


def compoile_unet(diffusion_model, graph_file):
    offload_device = model_management.unet_offload_device()
    load_device = model_management.get_torch_device()

    print(f" OneDiffCheckpointLoaderSimple load_checkpoint file_path {graph_file}")

    compile_options = OneflowCompileOptions()
    compile_options.graph_file = graph_file
    compile_options.graph_file_device = load_device
    diffusion_model = oneflow_compile(diffusion_model, options=compile_options)

    return diffusion_model


def quantize_unet(diffusion_model, calibrate_info, inplace=True):
    diffusion_model = quantize_model(
        model=diffusion_model, inplace=inplace, calibrate_info=calibrate_info
    )
    return diffusion_model



import base64
import io
import os
from itertools import product
from pathlib import Path
from typing import Any, Dict, Iterable, List

import numpy as np
import requests
from PIL import Image
from skimage.metrics import structural_similarity as ssim

WEBUI_SERVER_URL = "http://127.0.0.1:7860"
TXT2IMG_API_ENDPOINT = "sdapi/v1/txt2img"
IMG2IMG_API_ENDPOINT = "sdapi/v1/img2img"
OPTIONS_API_ENDPOINT = "sdapi/v1/options"
SEED = 1
NUM_STEPS = 20
HEIGHT = 1024
WIDTH = 1024
IMG2IMG_TARGET_FOLDER = "/share_nfs/onediff_ci/sd-webui/images/img2img"
TXT2IMG_TARGET_FOLDER = "/share_nfs/onediff_ci/sd-webui/images/txt2img"
SAVED_GRAPH_NAME = "saved_graph"

os.makedirs(IMG2IMG_TARGET_FOLDER, exist_ok=True)
os.makedirs(TXT2IMG_TARGET_FOLDER, exist_ok=True)


def get_base_args() -> Dict[str, Any]:
    return {
        "prompt": "1girl",
        "negative_prompt": "",
        "seed": SEED,
        "steps": NUM_STEPS,
        "width": WIDTH,
        "height": HEIGHT,
        "cfg_scale": 7,
        "n_iter": 1,
        "batch_size": 1,
        # Enable OneDiff speed up
        "script_name": "onediff_diffusion_model",
    }


def get_extra_args() -> List[Dict[str, Any]]:
    quant_args = [
        {
            "script_args": [
                x,  # quantization
                None,  # graph_checkpoint
                "",  # saved_graph_name
            ]
        }
        for x in [True, False]
    ]

    txt2img_args = [
        {},
        {"init_images": [get_init_image()]},
    ]

    return [
        quant_args,
        txt2img_args,
    ]


def get_all_args() -> Iterable[Dict[str, Any]]:
    for extra_args in product(*get_extra_args()):
        args = get_base_args()
        for extra_arg in extra_args:
            args = {**args, **extra_arg}
        yield args


def is_txt2img(data: Dict[str, Any]) -> bool:
    return "init_images" not in data


def is_quant(data: Dict[str, Any]) -> bool:
    return data["script_args"][0]


def encode_file_to_base64(path: str) -> str:
    with open(path, "rb") as file:
        return base64.b64encode(file.read()).decode("utf-8")


def post_request_and_check(url: str, data: Dict[str, Any]):
    response = requests.post(url, json=data)
    assert response.status_code == 200
    return response


def get_image_byte_from_response(response):
    return base64.b64decode(response.json()["images"][0])


def get_image_array_from_response(response):
    imgdata = base64.b64decode(response.json()["images"][0])
    return np.array(Image.open(io.BytesIO(imgdata)))


def get_init_image():
    img_path = str(Path(__file__).parent / "cat.png")
    return encode_file_to_base64(img_path)


def cal_ssim(src: np.ndarray, generated: np.ndarray) -> float:
    ssim_score = ssim(src, generated, multichannel=True, win_size=3)
    return ssim_score


def generate_image(filename: str, data: Dict[str, Any]):
    endpoint = TXT2IMG_API_ENDPOINT if is_txt2img(data) else IMG2IMG_API_ENDPOINT
    url = f"{WEBUI_SERVER_URL}/{endpoint}"
    response = post_request_and_check(url, data)
    image = get_image_byte_from_response(response)
    with open(filename, "wb") as file:
        file.write(image)


def get_target_image_filename(data: Dict[str, Any]) -> str:
    parent_path = TXT2IMG_TARGET_FOLDER if is_txt2img(data) else IMG2IMG_TARGET_FOLDER
    if not Path(parent_path).exists():
        Path(parent_path).mkdir(mode=777, parents=True)

    txt2img_str = "txt2img" if is_txt2img(data) else "img2img"
    quant_str = "-quant" if is_quant(data) else ""
    return f"{parent_path}/onediff{quant_str}-{txt2img_str}-w{WIDTH}-h{HEIGHT}-seed-{SEED}-numstep-{NUM_STEPS}.png"


def check_and_generate_images():
    for data in get_all_args():
        image_path = get_target_image_filename(data)
        if not Path(image_path).exists():
            print(f"Generating image for {get_data_summary(data)}...")
            generate_image(image_path, data)
        else:
            print(f"Image for {get_data_summary(data)} exists, skip generating...")


def get_data_summary(data: Dict[str, Any]) -> Dict[str, bool]:
    return {
        "is_txt2img": is_txt2img(data),
        "is_quant": is_quant(data),
    }


def dump_image(src_img: np.ndarray, target_img: np.ndarray, filename: str):
    combined_img = np.concatenate((src_img, target_img), axis=1)
    image = Image.fromarray(combined_img)
    image.save(f"{filename}.png")


def get_threshold(data: Dict[str, Any]):
    if is_quant(data):
        return 0.7
    else:
        return 0.95



# ComfyUI/comfy/hijack_model_management.py
import oneflow as flow  # usort: skip
from comfy.model_management import soft_empty_cache

from ..sd_hijack_utils import Hijacker


def hijack_soft_empty_cache(original_func, *args, **kwargs):
    original_func(*args, **kwargs)
    if flow.cuda.is_available():
        flow.cuda.empty_cache()


# Hijacker
model_management_hijacker = Hijacker()

model_management_hijacker.register(
    soft_empty_cache, hijack_soft_empty_cache, lambda *args, **kwags: True
)



"""oneflow_compiled UNetModel"""


from pathlib import Path

import gradio as gr
import modules.scripts as scripts
import modules.sd_models as sd_models
import modules.shared as shared
import onediff_controlnet
import onediff_shared
from compile import (
    get_compiled_graph,
    get_onediff_backend,
    OneDiffBackend,
    SD21CompileCtx,
    VaeCompileCtx,
)
from compile.nexfort.utils import add_nexfort_optimizer
from modules import script_callbacks
from modules.processing import process_images
from modules.ui_common import create_refresh_button

from onediff.utils import logger, parse_boolean_from_env
from onediff_hijack import do_hijack as onediff_do_hijack
from onediff_lora import HijackLoraActivate

# from onediff.optimization.quant_optimizer import varify_can_use_quantization
from onediff_utils import (
    check_structure_change,
    get_all_compiler_caches,
    hints_message,
    load_graph,
    onediff_enabled_decorator,
    onediff_gc,
    refresh_all_compiler_caches,
    save_graph,
    varify_can_use_quantization,
)


class UnetCompileCtx(object):
    """The unet model is stored in a global variable.
    The global variables need to be replaced with compiled_unet before process_images is run,
    and then the original model restored so that subsequent reasoning with onediff disabled meets expectations.
    """

    def __init__(self, enabled):
        self.enabled = enabled

    def __enter__(self):
        if not self.enabled:
            return
        self._original_model = shared.sd_model.model.diffusion_model
        shared.sd_model.model.diffusion_model = (
            onediff_shared.current_unet_graph.graph_module
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.enabled:
            return
        shared.sd_model.model.diffusion_model = self._original_model


class Script(scripts.Script):
    def title(self):
        return "onediff_diffusion_model"

    def ui(self, is_img2img):
        with gr.Row():
            # TODO: set choices as Tuple[str, str] after the version of gradio specified webui upgrades
            compiler_cache = gr.Dropdown(
                label="Compiler caches (Beta)",
                choices=["None"] + get_all_compiler_caches(),
                value="None",
                elem_id="onediff_compiler_cache",
            )
            create_refresh_button(
                compiler_cache,
                refresh_all_compiler_caches,
                lambda: {"choices": ["None"] + get_all_compiler_caches()},
                "onediff_refresh_compiler_caches",
            )
            save_cache_name = gr.Textbox(label="Saved cache name (Beta)")
        with gr.Row():
            always_recompile = gr.components.Checkbox(
                label="always_recompile",
                visible=parse_boolean_from_env("ONEDIFF_DEBUG"),
            )
        gr.HTML(
            hints_message,
            elem_id="hintMessage",
            visible=not varify_can_use_quantization(),
        )
        is_quantized = gr.components.Checkbox(
            label="Model Quantization(int8) Speed Up",
            visible=varify_can_use_quantization(),
        )
        return [is_quantized, compiler_cache, save_cache_name, always_recompile]

    def show(self, is_img2img):
        return True

    @onediff_enabled_decorator
    @onediff_controlnet.onediff_controlnet_decorator
    def run(
        self,
        p,
        quantization=False,
        compiler_cache=None,
        saved_cache_name="",
        always_recompile=False,
        backend=None,
    ):
        # restore checkpoint_info from refiner to base model if necessary
        if (
            sd_models.checkpoint_aliases.get(
                p.override_settings.get("sd_model_checkpoint")
            )
            is None
        ):
            p.override_settings.pop("sd_model_checkpoint", None)
            sd_models.reload_model_weights()
            onediff_gc()

        backend = backend or get_onediff_backend()
        current_checkpoint_name = shared.sd_model.sd_checkpoint_info.name
        ckpt_changed = (
            shared.sd_model.sd_checkpoint_info.name
            != onediff_shared.current_unet_graph.name
        )
        structure_changed = check_structure_change(
            onediff_shared.previous_unet_type, shared.sd_model
        )
        quantization_changed = (
            quantization != onediff_shared.current_unet_graph.quantized
        )
        controlnet_enabled_status_changed = (
            onediff_shared.controlnet_enabled != onediff_shared.previous_is_controlnet
        )
        need_recompile = (
            (
                quantization and ckpt_changed
            )  # always recompile when switching ckpt with 'int8 speed model' enabled
            or structure_changed  # always recompile when switching model to another structure
            or quantization_changed  # always recompile when switching model from non-quantized to quantized (and vice versa)
            or controlnet_enabled_status_changed
            or always_recompile
        )
        if need_recompile:
            if not onediff_shared.controlnet_enabled:
                onediff_shared.current_unet_graph = get_compiled_graph(
                    shared.sd_model,
                    quantization=quantization,
                    backend=backend,
                )
                load_graph(onediff_shared.current_unet_graph, compiler_cache)
        else:
            logger.info(
                f"Model {current_checkpoint_name} has same sd type of graph type {onediff_shared.previous_unet_type}, skip compile"
            )

        with UnetCompileCtx(not onediff_shared.controlnet_enabled), VaeCompileCtx(
            backend=backend
        ), SD21CompileCtx(), HijackLoraActivate():
            proc = process_images(p)
        save_graph(onediff_shared.current_unet_graph, saved_cache_name)

        return proc


def on_ui_settings():
    section = ("onediff", "OneDiff")
    shared.opts.add_option(
        "onediff_compiler_caches_path",
        shared.OptionInfo(
            str(Path(__file__).parent.parent / "compiler_caches"),
            "Directory for onediff compiler caches",
            section=section,
        ),
    )
    shared.opts.add_option(
        "onediff_compiler_backend",
        shared.OptionInfo(
            "oneflow",
            "Backend for onediff compiler (if you switch backend, you need to restart webui service)",
            gr.Radio,
            {"choices": [OneDiffBackend.ONEFLOW, OneDiffBackend.NEXFORT]},
            section=section,
        ),
    )


def cfg_denoisers_callback(params):
    pass


script_callbacks.on_ui_settings(on_ui_settings)
# script_callbacks.on_cfg_denoiser(cfg_denoisers_callback)
onediff_do_hijack()


script_callbacks.on_list_optimizers(add_nexfort_optimizer)



from pathlib import Path

import comfy
from comfy.ldm.modules.attention import attention_pytorch
from comfy.ldm.modules.diffusionmodules.model import AttnBlock

from .attention import attention_pytorch_oneflow


from nodes import *  # must imported before import comfy
from onediff.infer_compiler.backends.oneflow.transform import register
from onediff.infer_compiler.backends.oneflow.utils.version_util import (
    is_community_version,
)

from .attention import (
    CrossAttention as CrossAttention1f,
    SpatialTransformer as SpatialTransformer1f,
    SpatialVideoTransformer as SpatialVideoTransformer1f,
)
from .deep_cache_unet import DeepCacheUNet, FastDeepCacheUNet
from .linear import Linear as Linear1f
from .util import AlphaBlender as AlphaBlender1f
from .vae_patch import AttnBlock as AttnBlock1f

if hasattr(comfy.ops, "disable_weight_init"):
    comfy_ops_Linear = comfy.ops.disable_weight_init.Linear
else:
    print(
        "Warning: ComfyUI version is too old, please upgrade it. github: git@github.com:comfyanonymous/ComfyUI.git "
    )
    comfy_ops_Linear = comfy.ops.Linear

torch2of_class_map = {
    comfy.ldm.modules.attention.CrossAttention: CrossAttention1f,
    attention_pytorch: attention_pytorch_oneflow,
    comfy.ldm.modules.attention.SpatialTransformer: SpatialTransformer1f,
    comfy.ldm.modules.attention.SpatialVideoTransformer: SpatialVideoTransformer1f,
    comfy.ldm.modules.diffusionmodules.util.AlphaBlender: AlphaBlender1f,
    comfy_ops_Linear: Linear1f,
    AttnBlock: AttnBlock1f,
}

from .openaimodel import (
    UNetModel as UNetModel1f,
    Upsample as Upsample1f,
    VideoResBlock as VideoResBlock1f,
)

torch2of_class_map.update(
    {
        comfy.ldm.modules.diffusionmodules.openaimodel.Upsample: Upsample1f,
        comfy.ldm.modules.diffusionmodules.openaimodel.UNetModel: UNetModel1f,
        comfy.ldm.modules.diffusionmodules.openaimodel.VideoResBlock: VideoResBlock1f,
    }
)

register(torch2oflow_class_map=torch2of_class_map)



import comfy
import oneflow as flow  # usort: skip
import torch
from comfy.controlnet import ControlLora, ControlLoraOps, ControlNet
from onediff.infer_compiler import oneflow_compile

__all__ = ["OneDiffControlLora"]


def set_attr_of(obj, attr, value):
    def _set_attr_of(obj, attr, value):
        obj = obj._deployable_module_model._oneflow_module
        value = flow.utils.tensor.from_torch(value)
        attrs = attr.split(".")
        for name in attrs[:-1]:
            obj = getattr(obj, name)
        prev = getattr(obj, attrs[-1])
        prev.copy_(value)

    exist_oneflow_module = (
        getattr(obj._deployable_module_model, "_oneflow_module", None) is not None
    )

    if exist_oneflow_module:
        _set_attr_of(obj, attr, value)
    else:
        attrs = attr.split(".")
        for name in attrs[:-1]:
            obj = getattr(obj, name)
        prev = getattr(obj, attrs[-1])
        setattr(obj, attrs[-1], torch.nn.Parameter(value, requires_grad=False))
        del prev


class OneDiffControlLora(ControlLora):
    @classmethod
    def from_controllora(
        cls, controlnet: ControlLora, *, gen_compile_options: callable = None
    ):
        c = cls(
            controlnet.control_weights,
            global_average_pooling=controlnet.global_average_pooling,
        )
        controlnet.copy_to(c)
        c._oneflow_model = None
        c.gen_compile_options = gen_compile_options
        return c

    def pre_run(self, model, percent_to_timestep_function):
        dtype = model.get_dtype()
        # super().pre_run(model, percent_to_timestep_function)
        ControlNet.pre_run(self, model, percent_to_timestep_function)
        self.manual_cast_dtype = model.manual_cast_dtype

        if self._oneflow_model is None:
            controlnet_config = model.model_config.unet_config.copy()
            controlnet_config.pop("out_channels")
            controlnet_config["hint_channels"] = self.control_weights[
                "input_hint_block.0.weight"
            ].shape[1]
            dtype = model.get_dtype()

            if self.manual_cast_dtype is None:

                class control_lora_ops(ControlLoraOps, comfy.ops.disable_weight_init):
                    pass

            else:

                class control_lora_ops(ControlLoraOps, comfy.ops.manual_cast):
                    pass

                dtype = self.manual_cast_dtype

            controlnet_config["operations"] = control_lora_ops()
            self.control_model = comfy.cldm.cldm.ControlNet(**controlnet_config)
            self.control_model.to(dtype)
            self.control_model.to(comfy.model_management.get_torch_device())

            file_device_dict = (
                self.gen_compile_options(self.control_model)
                if self.gen_compile_options is not None
                else {}
            )
            self._oneflow_model = oneflow_compile(self.control_model)
            compiled_options = self._oneflow_model._deployable_module_options
            compiled_options.graph_file = file_device_dict.get("graph_file", None)
            compiled_options.graph_file_device = file_device_dict.get(
                "graph_file_device", None
            )

        self.control_model = self._oneflow_model

        diffusion_model = model.diffusion_model
        sd = diffusion_model.state_dict()
        cm = self.control_model.state_dict()
        for k in sd:
            # comfy.model_management.resolve_lowvram_weight(weight, model, key) always returns weight
            # and is removed now
            # weight = comfy.model_management.resolve_lowvram_weight(
            #     sd[k], diffusion_model, k
            # )
            weight = sd[k]
            try:
                set_attr_of(self.control_model, k, weight)
            except Exception as e:
                pass

        for k in self.control_weights:
            if k not in {"lora_controlnet"}:
                weight = (
                    self.control_weights[k]
                    .to(dtype)
                    .to(comfy.model_management.get_torch_device())
                )
                set_attr_of(self.control_model, k, weight)

    def copy(self):
        c = OneDiffControlLora(
            self.control_weights,
            global_average_pooling=self.global_average_pooling,
        )
        self.copy_to(c)
        c._oneflow_model = self._oneflow_model
        c.gen_compile_options = self.gen_compile_options
        return c



"""
Torch run example: python examples/text_to_image_sdxl_turbo.py
Compile to oneflow graph example: python examples/text_to_image_sdxl_turbo.py --compile
"""
import argparse
import os
import time

import torch
import oneflow as flow  # usort: skip

from diffusers import AutoPipelineForText2Image
from onediff.infer_compiler import oneflow_compile

parser = argparse.ArgumentParser()
parser.add_argument("--base", type=str, default="stabilityai/sdxl-turbo")
parser.add_argument("--variant", type=str, default="fp16")
parser.add_argument(
    "--prompt",
    type=str,
    default="A cinematic shot of a baby racoon wearing an intricate italian priest robe.",
)
parser.add_argument("--height", type=int, default=512)
parser.add_argument("--width", type=int, default=512)
parser.add_argument("--n_steps", type=int, default=4)
parser.add_argument("--saved_image", type=str, required=False, default="sdxl-out.png")
parser.add_argument("--warmup", type=int, default=1)
parser.add_argument("--seed", type=int, default=1)
parser.add_argument(
    "--compile", type=(lambda x: str(x).lower() in ["true", "1", "yes"]), default=True
)
args = parser.parse_args()

# Normal SDXL turbo pipeline init.
OUTPUT_TYPE = "pil"

# SDXL turbo base: AutoPipelineForText2Image
base = AutoPipelineForText2Image.from_pretrained(
    args.base,
    torch_dtype=torch.float16,
    variant=args.variant,
    use_safetensors=True,
)
base.to("cuda")

# Compile unet with oneflow
if args.compile:
    print("unet is compiled to oneflow.")
    base.unet = oneflow_compile(base.unet)
    print("vae is compiled to oneflow.")
    base.vae.decoder = oneflow_compile(base.vae.decoder)

# Warmup
for i in range(args.warmup):
    torch.manual_seed(args.seed)
    image = base(
        prompt=args.prompt,
        height=args.height,
        width=args.width,
        num_inference_steps=args.n_steps,
        guidance_scale=0.0,
        output_type=OUTPUT_TYPE,
    ).images

# Normal SDXL turbo run
torch.manual_seed(args.seed)
start_t = time.time()

image = base(
    prompt=args.prompt,
    height=args.height,
    width=args.width,
    num_inference_steps=args.n_steps,
    guidance_scale=0.0,
    output_type=OUTPUT_TYPE,
).images
end_t = time.time()
print(f"e2e ({args.n_steps} steps) elapsed: {end_t - start_t} s")

image[0].save(f"h{args.height}-w{args.width}-{args.saved_image}")
